﻿// SpringSim class - Actionscript 3 Version - Jim Bumgardner
//
// A class for providing spring motion to movie clips.
//
// If you're curious about the physics, check out the function 'doSimulate'
// below, it is heavily commented.  Everything else here is meant to support
// that function.
//

package {
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.display.DisplayObject;
	import flash.utils.getTimer;

	public class SpringSim
	{
		static const kDegToRad = Math.PI/180;
		static const RENDER_SPRINGS 	= 1;
		static const RENDER_DOTS 		= 2;
		static const RENDER_CURVY 		= 4;
		static const RENDER_POLYS 		= 8;
		static const dotRad = 5;
		
		var parMC:Sprite;
		
		var gravK, stiffK, dampK, renderFlags, simFloor, selectedDot,
			autoReverse, springThickness, floorDamping, springColor, springAlpha,
			editMode, timeStart, timeSnap, lastReverse, isDraggable;


		var isOverDot:Boolean = false;
		var dragDot;

		// muscle stuff
		var waveAmp, wavePhase, waveSpeed, waveDir;
			
		var dots = [], springs = [], polys = [];
		var springBG;
		
		function SpringSim(parMC:Sprite, gravK:Number=1, stiffK:Number=.5, frictionK:Number=.1, isDraggable:Boolean=true)
		{
			this.isDraggable = isDraggable;
			this.parMC = parMC;
			this.gravK = gravK;
			this.stiffK = stiffK;
			this.dampK = 1 - frictionK;
			this.renderFlags = 0;
			this.simFloor = parMC.stage.stageHeight;
			this.selectedDot = undefined;
			this.autoReverse = true;
			this.springThickness = 2;
			this.floorDamping = 1;
			this.springColor = 0x000000;
			this.springAlpha = .5;
			this.editMode = 0;
			this.timeStart = getTimer();
			this.timeSnap = 0;
			this.lastReverse = -2;
			this.reset();
			this.setEditMode(0);
			springBG = new Sprite();
			parMC.addChild(springBG);
		}


		function setRenderFlags(f)
		{
			this.renderFlags = f;
			this.parMC.graphics.clear();
			var showDots = (renderFlags & RENDER_DOTS) != 0;
			for (var i = 0; i < dots.length; ++i) {
				dots[i].showDot = showDots;
				dots[i].drawDot();
			}
			drawSprings();
		
		}

		function reset()
		{
			trace('reset');
			var nbrDots = dots.length;
			for (var i = 0; i < nbrDots; ++i) {
				parMC.removeChild( dots[i] );
			}
			springs = [];
			dots = [];
			polys = [];
			drawSprings();
		}

		public function setDrawCurvy(s)
		{
			if (s)
				renderFlags |= RENDER_CURVY;
			else
				renderFlags &= ~RENDER_CURVY;
			setRenderFlags(renderFlags);
		}
		
		public function setGravity(gravK)
		{
			this.gravK = gravK;
		}
		
		public function setStiffness(stiffK)
		{
			this.stiffK = stiffK;
		}

		public function setFriction(frictionK)
		{
			this.dampK = 1-frictionK;
		}

		public function setFloorRoughness(frictionK)
		{
			this.floorDamping = 1-frictionK;
		}
		
		public function setWave(amp,phase,speed)
		{
			this.waveAmp = amp;
			this.wavePhase = phase;
			this.waveSpeed = speed;
			this.waveDir = 1;
			this.timeSnap = 0;
			this.timeStart = getTimer();
		}


		public function doReverse()
		{
			this.timeSnap = this.timeSnap + (getTimer() - this.timeStart)*this.waveDir;
			this.waveDir *= -1;
			this.timeStart = getTimer();
			for (var i = dots.length-1; i >= 0; --i)  
			{
				var mc = dots[i];
				mc.vx *= -1;
			}
		}
		
		public function setAutoReverse(t)
		{
			this.autoReverse = t;
		}
		
		public function addSpring(dot1, dot2, tension=1):Spring
		{
			registerDot(dot1);
			registerDot(dot2);
			var spr = new Spring(dot1, dot2, tension);
			springs.push(spr);
			drawSprings();
			return spr;
		}

		public function findSpring(dot1, dot2):Spring
		{
			for (var i = 0; i < springs.length; ++i) {
				var spr = springs[i];
				if (spr.dot1 == dot1 && spr.dot2 == dot2)
					return spr;
				if (spr.dot2 == dot1 && spr.dot1 == dot2)
					return spr;
			}
			return undefined;
		}

		public function addPolygon(pts, clr, curvy):SSPolygon
		{
			var sprs = [0];
			for (var i = 1; i < pts.length; ++i)
				sprs[i] = this.findSpring(pts[i-1], pts[i]);
			var poly = new SSPolygon(pts, clr, curvy, sprs);
			polys.push(poly);
			drawSprings();
			return poly;
		}

		public function toggleShowGuide(s)
		{
			if (s)
				renderFlags |= (RENDER_SPRINGS | RENDER_DOTS);
			else
				renderFlags &= ~(RENDER_SPRINGS | RENDER_DOTS);
			setRenderFlags(renderFlags);
			springBG.graphics.clear();
		}

		function drawSprings(evt=null)
		{
			// return;
			if (renderFlags == 0)
				return;
			var graf = springBG.graphics;
			graf.clear();
			var i,spr,mc1,mc2,dx,dy,cx,cy,ang,r;
			if (renderFlags & RENDER_SPRINGS)
			{
				graf.lineStyle(springThickness,springColor, springAlpha);
				if (editMode == 1 && dragDot != undefined)
				{
					for (i = springs.length-1; i >= 0; --i) 
					{
						spr = springs[i];
						if (spr.dot1 == dragDot ||
							spr.dot2 == dragDot) 
						{
							spr.dist = spr.restLen = spr.getLength();
						}
					}
				}
				for (i = springs.length-1; i >= 0; --i) 
				{
					spr = springs[i];
					mc1 = spr.dot1;
					mc2 = spr.dot2;
					graf.moveTo(mc1.x, mc1.y);
					if (!(renderFlags & RENDER_CURVY) || mc1.isAnchor || mc2.isAnchor)
						graf.lineTo(mc2.x, mc2.y);
					else
					{
						dx = mc2.x - mc1.x;
						dy = mc2.y - mc1.y;
						cx = (mc1.x + mc2.x)/2;
						cy = (mc1.y + mc2.y)/2;
						ang = Math.atan2(dy,dx)+Math.PI/2;
						r = spr.dist - spr.restLen;
						graf.curveTo(cx+Math.cos(ang)*r, cy+Math.sin(ang)*r,mc2.x, mc2.y);		
					}
				}
			}
			if (renderFlags & RENDER_POLYS)
			{
				var len = polys.length;
				for (i = 0; i < len; ++i) 
				{
					var poly = polys[i];
					graf.beginFill(poly.clr, 1);
					mc1 = poly.pts[0];
					graf.moveTo(mc1.x, mc1.y);
					for (var p = 1; p < poly.pts.length; ++p) {
						mc2 = poly.pts[p];
						if (!poly.curvy || poly.sprs[p] == undefined)
							graf.lineTo(mc2.x, mc2.y);
						else {
							spr = poly.sprs[p];
							dx = mc2.x - mc1.x;
							dy = mc2.y - mc1.y;
							cx = (mc1.x + mc2.x)/2;
							cy = (mc1.y + mc2.y)/2;
							ang = Math.atan2(dy,dx)+Math.PI/2;
							r = spr.dist - spr.restLen;
							graf.curveTo(cx+Math.cos(ang)*r, cy+Math.sin(ang)*r,mc2.x, mc2.y);		
						}
						mc1 = mc2;
					}
					graf.endFill();
				}
			}
		}

		// DOSIMULATE
		// THIS IS WHERE THE FUN PHYSICS ARE: RIGHT HERE IN THIS VERY FUNCTION
		//
		// Basically we compute the forces affecting each MC, and then
		// use them to tweak the motion
		//
		// Note: Unlike many physics code examples, we completely ignore
		// time, assuming that frame-rate is relatively constant (it is close enough).
		//
		// For our purposes this is just fine -it makes the code run faster
		// and makes it more readable.
		public function doSimulate(evt)
		{
			// return;
			var elapsed = timeSnap + (getTimer() - timeStart)*waveDir;
			var waveIdx = elapsed * 0.1 * Math.PI * waveSpeed + wavePhase;
			// trace("sim " + dots.length + ", " + springs.length);
			// zero out forces & add gravity
			var i,mc;
			for (i = dots.length-1; i >= 0; --i) 
			{
				mc = dots[i];
				mc.fx = 0;
				mc.fy = 0;
				if (!mc.isAnchor && !mc.isDragging)
					mc.fy = gravK;
			}
			// figure out the amount and direction of the force the springs are applying to the particles
			for (i = springs.length-1; i >= 0; --i) 
			{
				var spr = springs[i];
				var mc1 = spr.dot1;
				var mc2 = spr.dot2;
		
				// distance between spring endpoints
				var dist = spr.getLength();
				
				spr.dist = dist;
		
				var f1 = stiffK * (dist - spr.restLen);
				if (spr.isMuscle) 
					f1 += .5*spr.restLen*spr.amp*waveAmp*Math.sin(waveIdx+spr.phase);
				// spring force is proportional to difference with resting length
						// HERE BE HOOKE'S LAW
						// Robert Hooke - A great man indeed :)  Also drew a mean flea!
						// If only he had flash...
		
				// and is applied in direction of spring
				var spr_fx = f1 * (mc2.x - mc1.x)/dist;
				var spr_fy = f1 * (mc2.y - mc1.y)/dist;

				// springs only affect 'loose' MCs
				if (!mc1.isAnchor && !mc1.isDragging) 
				{
					mc1.fx += spr_fx;
					mc1.fy += spr_fy;
				}
				// spring has opposite affect on each connected MC
				if (!mc2.isAnchor && !mc2.isDragging) 
				{
					mc2.fx -= spr_fx;
					mc2.fy -= spr_fy
				}
			}
		
			// For each MC
			//
			var minx = parMC.stage.stageWidth;
			var maxx = 0;
			for (i = dots.length-1; i >= 0; --i)  
			{
				mc = dots[i];
				if (mc.isDragging)
					continue;
		
				// Applying damping (friction, viscosity - whatever) - this is what
				// slows us down
				mc.vx *= dampK;
				mc.vy *= dampK;

		
				// Apply spring forces and gravity to the MC's current velocity
				mc.vx += mc.fx;
				mc.vy += mc.fy;

				if (mc.y+mc.vy > simFloor) {
					mc.vy -= mc.y+mc.vy - simFloor;
					mc.vy *= floorDamping;
					mc.vx *= floorDamping;
				}
				// Use the velocity to adjust our position
				//
				mc.x += mc.vx;	
				mc.y += mc.vy;	
				maxx = Math.max(maxx, mc.x);
				minx = Math.min(minx, mc.x);
			}
			if ((maxx > parMC.stage.stageWidth && lastReverse != 1) || (minx < 10 && lastReverse != 0)) {
				lastReverse = (maxx > parMC.stage.stageWidth);
				doReverse();
			}
			// Redraw the springs (the guide), if necessary
			drawSprings();
		}

		// Editing functions - not used for basic spring sim stuff
		//
		
		// use this if you want SpringSim to provide the movieclips for the dots
		// you may also provide your own MCs
		function addDotMC(px,py,isFixed):MassSprite
		{
			var mc = new MassSprite(this);
			mc.x = px;
			mc.y = py;
			mc.setDotProps(isFixed, true);
			mc.showDot();
			registerDot(mc);
			parMC.addChild(mc);
			return mc;
		}

		public function hasMC(mc):Boolean
		{
			for (var i = dots.length-1; i >= 0; --i) 
			{
				if (dots[i] == mc)
					return true;
			}
			return false;
		}

		public function registerDot(mc)
		{
			if (!hasMC(mc))
			{
				trace('registering ' + mc);
				dots.push(mc);
				mc.drawDot();
				if (isDraggable)
				{
					mc.addEventListener(MouseEvent.MOUSE_OVER, mc.overDot );
					mc.addEventListener(MouseEvent.MOUSE_OUT, mc.outofDot );
					mc.addEventListener(MouseEvent.MOUSE_DOWN, mc.pressDot);
					mc.addEventListener(MouseEvent.MOUSE_UP, mc.releaseDot);
				}
			}
		}

		
		function addAMouseDot():MassSprite // !! phase out layer number tracking
		{
			if (parMC.mouseY < 300 && parMC.mouseY > 0 &&
				parMC.mouseX < parMC.stage.stageWidth && parMC.mouseX > 0)
			{
				var mc = addDotMC(parMC.mouseX, parMC.mouseY, false);
				if (selectedDot != undefined)
					addSpring(selectedDot, mc);
				doSelectDot(mc);
				mc.creationTime = getTimer();
				return mc;
			}
			return undefined;
		}
		
		// set state of whether dots are visible or not
		function showDots(s) // affects built-ins only
		{
			var nbrDots = dots.length;
			for (var i = 0; i < nbrDots; ++i) {
				dots[i].setDotProps(dots[i].isAnchor, s);
			}
		}
		
		function doSelectDot(dot)
		{
			if (selectedDot != undefined) {
				selectedDot.isSelected = false;
				selectedDot.drawDot();
			}
			if (dot != undefined) {
				selectedDot = dot;
				dot.isSelected = true;
				dot.drawDot();
			}
			else
				selectedDot = undefined;
		}
		
		function deleteDot(dot)
		{
			var i;
			for (i = this.springs.length-1; i >= 0; --i) {
				if (this.springs[i].dot1 == dot || this.springs[i].dot2 == dot)
					this.springs.splice(i,1);
			}
			for (i = dots.length-1; i >= 0; --i) {
				if (dots[i] == dot)
					dots.splice(i, 1);
			}
			parMC.removeChild( dot );
			drawSprings();
			doSelectDot(undefined);
		}
		
		function setEditMode(f)
		{
			// trace("edit mode = " + f);
			editMode = f;
			switch (editMode) {
			case 0:	doSelectDot(undefined);
					this.parMC.addEventListener(Event.ENTER_FRAME, doSimulate);
					break;
			case 1:	this.parMC.removeEventListener(Event.ENTER_FRAME, doSimulate);
					break;
			case 2:	this.parMC.removeEventListener(Event.ENTER_FRAME, doSimulate);
					break;
			}
		}






	} // end class
	
} // end package

